(*
    Baldur's Gate roll dice utility - Main window
    Copyright (c) 1999-2003,2004 by Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TForm1 = class(TForm)
    Button1: TButton;
    Label1: TLabel;
    Timer1: TTimer;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    LblRoll1: TLabel;
    LblRoll2: TLabel;
    LblRoll3: TLabel;
    LblRoll4: TLabel;
    LblRoll5: TLabel;
    LblRoll6: TLabel;
    Label7: TLabel;
    LblTotal: TLabel;
    Label8: TLabel;
    EdtRequired: TEdit;
    Label9: TLabel;
    Label10: TLabel;
    LblCounter: TLabel;
    chkUnknown19: TCheckBox;
    Label11: TLabel;
    Label12: TLabel;
    lblBest: TLabel;
    lblAverage: TLabel;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    HBaldur: HWnd;
    FRunning: boolean;
    RollCount: integer;
    BestRoll: integer;
    TotalRoll: integer;
    function Grab( X, Y, Width, Height: integer ): TBitmap;
    function FindBaldur: HWnd;
    procedure ActivateBaldur( W: HWnd );
    function AskForRollValue( I: integer; B: TBitmap ): integer;
    procedure OnHotKey( var M: TMessage ); message WM_HOTKEY;
    procedure SetRunning( Value: boolean );
    procedure Reroll;
    property Running: boolean read FRunning write SetRunning;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses Unit2;

{$R *.DFM}

procedure MoveMouseTo( X, Y: dword );
begin
  Mouse_Event( MOUSEEVENTF_ABSOLUTE or MOUSEEVENTF_MOVE,
    X * 65536 div GetSystemMetrics(SM_CXSCREEN),
    Y * 65536 div GetSystemMetrics(SM_CYSCREEN),
    0, 0 );
end;

procedure PressKey( K: byte );
var
  S: byte;
begin
  S := MapVirtualKey( K, 0 );
  Keybd_Event( K, S, 0, 0 );
  Keybd_Event( K, S, KEYEVENTF_KEYUP, 0 );
end;

function TForm1.FindBaldur: HWnd;
begin
  Result := FindWindow( 'ChitinClass', nil );
end;

procedure TForm1.ActivateBaldur;
begin
  if( W = 0 ) then Exit;
  ShowWindow( W, SW_NORMAL );
  SetForegroundWindow( W );
  SetWindowPos( W, HWND_TOP, 0, 0, 0, 0, (* SWP_NOMOVE or *) SWP_NOSIZE );
  SetWindowPos( Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE );
  Application.ProcessMessages;
end;

function TForm1.Grab;
var
  BM: TBitmap;
  DC: HDC;
begin
  BM := TBitmap.Create;
  BM.Width := Width;
  BM.Height := Height;
  DC := GetDC( 0 );
  BitBlt( BM.Canvas.Handle,
    0, 0, Width, Height,
    DC,
    X, Y,
    SRCCOPY );
  ReleaseDC( 0, DC );
  Result := BM;
end;

type
  TBitmapCode = array[1..9] of integer;
  PBitmapCode = ^TBitmapCode;
const
  MinRollValue = 3;
  MaxRollValue = 18;
var
  RollCode : array[MinRollValue..MaxRollValue] of PBitmapCode;

procedure LoadRollCodeData( FileName: string );
var
  F: System.Text;
  S: string;
  C: PBitmapCode;
  V: integer;
  I: integer;
begin
  AssignFile( F, Filename );
  Reset( F );
  while( not Eof( F ) ) do begin
    ReadLn( F, S );
    S := Trim( S );
    if( S = '' )or( S[1] = ';' ) then
      continue;
    V := StrToInt( Copy( S, 1, 2 ) );
    New( C );
    Delete( S, 1, 5 );
    for I:=1 to 9 do begin
      C^[I] := StrToInt( Copy( S, 1, 5 ) );
      Delete( S, 1, 6 );
    end;
    RollCode[ V ] := C;
  end;
  CloseFile( F );
end;

procedure SaveRollCodeData( FileName: string );
var
  F: System.Text;
  C: PBitmapCode;
  V: integer;
  I: integer;
begin
  AssignFile( F, Filename );
  ReWrite( F );
  for I:=MinRollValue to MaxRollValue do begin
    C := RollCode[ I ];
    if( C = nil ) then
      continue;
    if( I < 10 ) then Write( F, '0' );
    Write( F, I, ' = ' );
    for V:=1 to 9 do
      Write( F, '$', IntToHex( C^[V], 4 ), ' ' );
    WriteLn( F );
  end;
  CloseFile( F );
end;

procedure GetCodeOfBitmap( B: TBitmap; var O: TBitmapCode );
var
  C, I, X, Y: integer;
  W: integer;
begin
  (* Compute roll code *)
  for Y:=0 to 8 do begin
    W := 0;
    for X:=0 to B.Width-1 do begin
      C := B.Canvas.Pixels[X,Y];
      I := (C and $FF)*30 + ((C shr 8) and $FF)*59 + ((C shr 16) and $FF)*11;
      I := I div 100;
      W := W shl 1;
      if( I >= 128 ) then Inc( W );
    end;
    O[ Y+1 ] := W;
  end;
end;

function BitmapToValue( B: TBitmap ): integer;
var
  O: TBitmapCode;
  C, I: integer;
begin
  Result := -1;
  (* Compute roll code *)
  GetCodeOfBitmap( B, O );
  (* Match *)
  for I := MinRollValue to MaxRollValue do begin
    if( RollCode[I] = nil ) then
      continue;
    C := 1;
    while( C <= 9 ) do begin
      if( RollCode[I]^[C] <> O[C] ) then
        break;
      Inc( C );
      if( C > 9 ) then begin
        Result := I;
        Exit;
      end;
    end;
  end;
end;

function TForm1.AskForRollValue( I: integer; B: TBitmap ): integer;
var
  C: PBitmapCode;
begin
  if( Running ) then Running := false;
  Result := -1;
  SetForegroundWindow( Handle );
  Activate;
  FormAskValue.LblRollEntry.Caption := IntToStr( I );
  FormAskValue.CboRollValue.ItemIndex := 0;
  if( FormAskValue.ShowModal = mrOk ) then begin
    New( C );
    GetCodeOfBitmap( B, C^ );
    Result := FormAskValue.CboRollValue.ItemIndex+3;
    RollCode[ Result ] := C;
    SaveRollCodeData( ExtractFilePath( Application.ExeName ) + 'rolldata.txt' );
  end;
  ActivateBaldur( HBaldur );
end;

const
  RollPosX = 386;
  RollPosY : array[1..6] of integer = ( 112, 143, 174, 205, 236, 267 );

procedure TForm1.SetRunning( Value: boolean );
begin
  if( FRunning <> Value ) then begin
    FRunning := Value;
    if( FRunning ) then
      Button1.Caption := 'Stop'
    else
      Button1.Caption := 'Start';
    Label9.Caption := 'Alt-Shift-S to ' + Button1.Caption;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  Running := not Running;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  FRunning := false;
  Left := Screen.Width - Width - 16;
  HBaldur := FindBaldur;
  if( HBaldur = 0 ) then begin
    MessageBox( 0, 'Cannot find Baldur''s Gate window', 'Error', MB_OK );
    Application.Terminate;
    Exit;
  end;
  ActivateBaldur( HBaldur );
  LoadRollCodeData( ExtractFilePath( Application.ExeName ) + 'rolldata.txt' );
  RegisterHotKey( Handle, 1, MOD_ALT or MOD_SHIFT, Ord('S') );
  RollCount := 0;
  BestRoll := 0;
  TotalRoll := 0;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
const
  Counter : integer = 0;
var
  P: TPoint;
  B: TBitmap;
  I: integer;
  V: integer;
  T: integer;
  O: boolean;
begin
  if( not Running ) then begin
    Exit;
  end;
  (* Get total *)
  T := 0;
  for I:=1 to 6 do begin
    P.X := RollPosX;
    P.Y := RollPosY[ I ];
    Windows.ClientToScreen( HBaldur, P );
    B := Grab( P.X, P.Y, 16, 9 );
    V := BitmapToValue( B );
    if( V < 0 ) then begin
      if( chkUnknown19.Checked )and( I = 1 ) then begin
        V := 18
      end
      else begin
        O := Running;
        Running := false;
        V := AskForRollValue( I, B );
        if( V < 0 ) then Exit;
        Running := O;
      end;
    end;
    case I of
    1 : LblRoll1.Caption := IntToStr( V );
    2 : LblRoll2.Caption := IntToStr( V );
    3 : LblRoll3.Caption := IntToStr( V );
    4 : LblRoll4.Caption := IntToStr( V );
    5 : LblRoll5.Caption := IntToStr( V );
    6 : LblRoll6.Caption := IntToStr( V );
    end;
    if( V >= 0 ) then Inc( T, V );
    B.Free;
  end;
  LblTotal.Caption := IntToStr( T );
  Inc( RollCount );
  LblCounter.Caption := IntToStr( RollCount );
  if( T > BestRoll ) then begin
    BestRoll := T;
    lblBest.Caption := IntToStr( T );
  end;
  Inc( TotalRoll, T );
  if( (RollCount and 15) = 0 ) then
    lblAverage.Caption := IntToStr(TotalRoll div RollCount);
  (* *)
  if( T >= StrToInt( EdtRequired.Text ) ) then begin
    FRunning := false;
    SetForegroundWindow( Handle );
    Activate;
    MessageBox( 0, 'Required total matched!', 'Yeah', MB_OK );
  end
  else begin
    Inc( Counter );
    if( Counter > 100 ) then begin
      Counter := 0;
      Sleep( Random(100) );
    end;
    Reroll;
  end;
end;

procedure TForm1.Reroll;
var
  P: TPoint;
begin
  P.X := 420;
  P.Y := 310;
  Windows.ClientToScreen( HBaldur, P );
  MoveMouseTo( P.X, P.Y );
  Mouse_Event( MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0 );
  Mouse_Event( MOUSEEVENTF_LEFTUP, 0, 0, 0, 0 );
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  UnregisterHotKey( Handle, 1 );
end;

procedure TForm1.OnHotKey( var M: TMessage );
begin
  Button1Click( Self );
end;

end.
